@@ -26,6 +26,15 @@
  * -remove dependence on TclX by #include tcl.h, tclUnix.h 
  * -change malloc/free to challoc/ckfree in case TCL_MEM_DEBUG
  *-----------------------------------------------------------------------------
+ * Version 2.1 February, 1994
+ * Tom Poindexter, Boulder Colorado
+ * tpoind@advtech.uswest.com  or tpoindex@nyx.cs.du.edu   
+ * -change Oratcl_Init to int and return ok
+ * -add "colprec" and "colscales" to oracols (thanks-Dan R. Schenck)
+ * -add optional repeat commands to orafetch command
+ * -don't call get_lda_err from failed OraLogon, may not have good lda
+ * -change way cur handles are made, don't use argv[1] anymore
+ *-----------------------------------------------------------------------------
  *
  */
 
@@ -168,7 +177,7 @@ int osql3 _ANSI_ARGS_((struct cda_def *, char *, int));
 #define ORATCLCURS      25	/* default number of curs available */
 #define ORA_HOST_AREA	256	/* Oracle host data area length */
 #define ORA_BUFF_SIZE	4096	/* conversion buffer size for various needs*/
-#define ORA_MSG_SIZE	502	/* oracle error message max size*/
+#define ORA_MSG_SIZE	1000	/* oracle error message max size*/
 #define ORA_CACHE       10      /* number of rows to fetch each time */
 
 /* parse_cols defines */
@@ -260,6 +269,9 @@ typedef struct OraTclCurs {	/* struct for cursor entries */
 static OraTclLdas   OraLdas[ORATCLLDAS];  
 static OraTclCurs   OraCurs[ORATCLCURS];  
 
+/* quick fetch flag */
+static int  fetch_ok;
+
 static char *OraHandlePrefix = "oratcl";  /* prefix used to identify handles*/
 /* lda handles are:  prefix , lda index          e.g.  oratcl0    */
 /* cur handles are:  lda handle , '.', cur index e.g.  oratcl0.0  */
@@ -453,8 +465,11 @@ clear_msg(interp)
     Tcl_SetVar2(interp, OraMsgArray, "rows",     "", TCL_GLOBAL_ONLY);
 
     /* index "collengths" & "coltypes" only meaningful after oracols command */
+    /* same for "colprecs" and "colscales"                                   */
     Tcl_SetVar2(interp, OraMsgArray, "collengths","", TCL_GLOBAL_ONLY);
     Tcl_SetVar2(interp, OraMsgArray, "coltypes",  "", TCL_GLOBAL_ONLY);
+    Tcl_SetVar2(interp, OraMsgArray, "colprecs",  "", TCL_GLOBAL_ONLY);
+    Tcl_SetVar2(interp, OraMsgArray, "colscales", "", TCL_GLOBAL_ONLY);
 
 
 }
@@ -788,6 +803,8 @@ append_cols (interp, hand)
     Tcl_SetVar2(interp, OraMsgArray, "rows", buf, TCL_GLOBAL_ONLY);
     Tcl_SetVar2(interp, OraMsgArray, "rc",   "0", TCL_GLOBAL_ONLY);
 
+    fetch_ok = 1;
+
 }
 
 
@@ -829,7 +846,7 @@ Tcl_KillOra (clientData)
  *   Tcl_CreateExtendedInterp.
  */
 
-void
+int
 Oratcl_Init (interp)
     Tcl_Interp *interp;
 {
@@ -909,6 +926,7 @@ Oratcl_Init (interp)
     /* other indices - correspond to error and message handler arguments */
     clear_msg(interp);
 
+    return TCL_OK;
 }
 
 
@@ -985,7 +1003,13 @@ Tcl_OraLogon (clientData, interp, argc, argv)
 
     if (orlon(OraLdas[hand].lda,OraLdas[hand].hda,argv[1], -1,NULL,-1) != 0) {
 	/* logon failed, cleanup */
-	get_lda_err(interp,hand);
+	/* can't use get_lda_err(), may not have connected to server*/
+        sprintf(buf,"%d",OraLdas[hand].lda->rc);
+        Tcl_SetVar2(interp,OraMsgArray,"rc",buf,TCL_GLOBAL_ONLY);
+        Tcl_SetVar2(interp,OraMsgArray,"errortxt",
+           "logon failed: invalid id/password or unable to connect to server",
+	    TCL_GLOBAL_ONLY);
+
         ckfree(OraLdas[hand].lda);
         ckfree(OraLdas[hand].hda);
 	Tcl_AppendResult (interp, argv[0], ": oralogon failed in orlon",
@@ -1163,7 +1187,7 @@ Tcl_OraOpen (clientData, interp, argc, argv)
 
 
     /* construct and return a cursor handle */
-    sprintf(buf,"%s.%d",argv[1],hand);   /* argv[1] is lda_handle */
+    sprintf(buf,"%s%d.%d",OraHandlePrefix,lda_hand,hand);  
 
     Tcl_SetVar2(interp, OraMsgArray, "handle",      buf, TCL_GLOBAL_ONLY);
     clear_msg(interp);
@@ -1246,6 +1270,9 @@ Tcl_OraSql (clientData, interp, argc, argv)
 	return TCL_ERROR;
     }
 
+    /* cancel any pending usage of cursor */
+    ocan(OraCurs[hand].cda);
+
     /* set up sql for execution */
     oparse(OraCurs[hand].cda, argv[2], ((long) -1), DEF_FLAG, LNG_FLAG);
 
@@ -1323,12 +1350,21 @@ Tcl_OraFetch (clientData, interp, argc, argv)
     long     old_rpc;
     ColBufs *col;
     int      i;
+    int      Tcl_OraFetchAll();
 
 
-    if ((hand = ora_prologue(interp, argc, argv, 2, " cur_handle")) == -1) {
+    if ((hand = ora_prologue(interp, argc, argv, 2, 
+	 " cur_handle  ?commands?")) == -1) {
 	return TCL_ERROR;
     }
 
+    /* if there's a third argument, let next proc do the work */
+    if (argc >= 3) {
+	return (Tcl_OraFetchAll(clientData, interp, argc, argv));
+    }
+
+    fetch_ok = 0;
+
     /* check if already exahusted */
 
     if (OraCurs[hand].fetch_end && 
@@ -1391,6 +1427,227 @@ Tcl_OraFetch (clientData, interp, argc, argv)
 
 
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Tcl_OraFetchAll --
+ *    Implements the orafetch with iterative commands:
+ *    usage: orafetch cur_handle tcl_stmts
+ *	                
+ *    results:
+ *	fetch all rows from existing query, execute tcl_stmt on each row
+ *	sets message array element "rc" with value of final fetch rcode
+ *      TCL_OK - handle is opened
+ *      TCL_ERROR - wrong # args, or handle not opened, tcl_stmt failed,
+ *                  or wrong number of columns
+ */
+
+static int
+Tcl_OraFetchAll (clientData, interp, argc, argv)
+    ClientData   clientData;
+    Tcl_Interp  *interp;
+    int          argc;
+    char       **argv;
+{
+    int      hand;
+    char     buf[ORA_MSG_SIZE];
+    char    *eval_ptr;
+    int      colsCnt;
+    char   **colsPtr;
+    int      i;
+    char    *s;
+    char    *s2;
+    char    *p;
+    ColBufs *col;
+    int      max_col = 0;
+
+#define NUMSTRS 300		/* number of individual strings to allow*/
+#define SUBCHAR '@'		/* substitution character */
+
+    int      num_str = 0;	/* number of strings to concat */
+    char    *str[NUMSTRS];      /* strings to pass to concat   */
+
+    int      colnum  = 0; 	/* number of colstr's used */
+
+    struct {			
+      int      column; 		/* column number index */
+      int      strnum;          /* str array index */
+    } colstr[NUMSTRS];
+
+    Tcl_DString evalStr;	/* tcl dynamic string for building eval */
+    int        inum;
+    int        icol;
+    int        tcl_rc;
+
+
+    if ((hand = ora_prologue(interp, argc, argv, 3, 
+			    " cur_handle  commands")) == -1) {
+	return TCL_ERROR;
+    }
+
+    /* check if already exahusted or nothing to fetch */
+
+    if (OraCurs[hand].fetch_end && 
+	OraCurs[hand].fetch_cnt >= OraCurs[hand].cda->rpc &&
+	OraCurs[hand].fetch_cnt >  0) {
+
+	sprintf(buf,"%d",NO_DATA_FOUND);
+        Tcl_SetVar2(interp, OraMsgArray, "rc",   buf, TCL_GLOBAL_ONLY);
+        sprintf(buf,"%ld",OraCurs[hand].cda->rpc);
+        Tcl_SetVar2(interp, OraMsgArray, "rows", buf, TCL_GLOBAL_ONLY);
+	return TCL_OK;
+    }
+
+    if (OraCurs[hand].col_list == NULL) {   
+	return TCL_OK;			/* nothing to fetch */
+    }
+
+    /* parse tcl_stmt for all '@' substitutions */
+
+    s  = argv[2];			/* start of tcl_stmt */
+    s2 = s;
+
+    while ( (p = strchr(s2,SUBCHAR)) != NULL) {
+
+      if (isdigit(*(p+1))) {		/* it's a substituion value ! */
+
+	  if (s != p) {			/* subchar not at the front of string?*/
+	      str[num_str] = s;		/* save the current front of string */
+	      *p = '\0';  		/* terminate the literal string     */
+	      num_str++;		/* into str array and               */
+	  }
+
+          i = atoi(p + 1);		/* the column number */
+
+	  max_col = (i > max_col) ? i : max_col;/* sanity check, max col num */
+
+	  colstr[colnum].column = i;		/* save column number */
+	  colstr[colnum].strnum = num_str;	/* and str positiion  */
+	  str[num_str] = "";			/* sanity check       */
+	  colnum++;
+	  num_str++;
+	  p++;				/* skip past col number */
+	  while (isdigit(*p)) {
+	      p++;
+	  }
+	  s  = p;			/* next point to look for subchar */
+	  s2 = p;			
+
+
+      } else {				/* it's a subchar without number */
+	  s2 = p + 1;			/* ignore and keep searching     */
+      }
+
+    }
+
+    if (num_str == 0) {    		/* check if no substitutions */
+	str[num_str] = argv[2];	
+	num_str = 1;	
+    } else {		
+	if (strlen(s) > 0) {		/* remainder of tcl_stmt string */
+            str[num_str] = s;
+	    num_str++;
+	}
+    }
+
+
+    /* make sure there are enough columns in result */
+    i = 0;
+    for (col = OraCurs[hand].col_list; col != NULL; col = col->next_buf) {
+	i++;
+    }
+    if (max_col > i) {
+	Tcl_AppendResult (interp, argv[0], ": @column number execeeds results", 
+			  (char *) NULL);
+	return TCL_ERROR;
+    }
+
+
+    /* loop until fetch exhausted */
+
+    if (Tcl_OraFetch(clientData, interp, 2, argv) == TCL_ERROR) {
+	Tcl_AppendResult (interp, ": ", argv[0], ": orafetch failed", 
+			  (char *) NULL);
+	return TCL_ERROR;
+    }
+
+    while (fetch_ok) {
+
+        /* split the result columns left in interp->result */
+	if (Tcl_SplitList(interp,interp->result,&colsCnt,&colsPtr) == 
+							    TCL_ERROR) {
+	   Tcl_AppendResult (interp, argv[0], ": split columns failed", 
+			      (char *) NULL);
+	   return TCL_ERROR;
+	}
+
+        /* build eval string from literals and columns */
+        Tcl_DStringInit(&evalStr);
+	eval_ptr = "";
+	for (inum=0, icol=0; inum < num_str; inum++) {
+	    if (inum == colstr[icol].strnum) {
+	        eval_ptr = Tcl_DStringAppend(&evalStr, " ", -1);
+		if (colstr[icol].column == 0) {    /* col 0 is entire result */
+		    eval_ptr = Tcl_DStringAppendElement(&evalStr,
+							interp->result);
+                } else {
+		    if (colstr[icol].column > colsCnt) {/* another sanity chk*/
+			Tcl_AppendResult (interp, argv[0], 
+			       ": column sanity failed!",(char *) NULL);
+	                Tcl_DStringFree(&evalStr); 	/* free eval string */
+	                ckfree((char *) colsPtr); 	/* free split array */
+			return TCL_ERROR;
+		    }
+		    eval_ptr = Tcl_DStringAppendElement(&evalStr,
+					  colsPtr[colstr[icol].column - 1]);
+		}
+	        eval_ptr = Tcl_DStringAppend(&evalStr, " ", -1);
+	        icol++;
+	    } else {
+	        eval_ptr = Tcl_DStringAppend(&evalStr, str[inum], -1);
+	    }
+	}
+
+        tcl_rc = Tcl_Eval(interp, eval_ptr);	/* do it! */
+
+	Tcl_DStringFree(&evalStr); 	/* free eval string */
+
+	ckfree((char *) colsPtr); 	/* free split array */
+
+	switch (tcl_rc) {		/* check on eval return code */
+	  case TCL_ERROR:
+	    Tcl_AppendResult (interp, ": eval failed while in ", argv[0], 
+			      (char *) NULL);
+	    return TCL_ERROR;
+	    break;
+
+	  case TCL_BREAK:		/* return sooner */
+	    return TCL_OK;
+	    break;
+
+	  default:
+	    break;
+	}
+
+	Tcl_ResetResult(interp);	/* reset interp result for next */
+ 
+        /* next fetch */
+        if (Tcl_OraFetch(clientData, interp, 2, argv) == TCL_ERROR) {
+	    Tcl_AppendResult (interp, ": ", argv[0], ": orafetch failed", 
+			      (char *) NULL);
+	    return TCL_ERROR;
+        }
+
+    }
+
+    return TCL_OK;
+
+}
+
+
+
+
 /*
  *----------------------------------------------------------------------
  *
@@ -1417,6 +1674,8 @@ Tcl_OraCols (clientData, interp, argc, argv)
     char   *type_ptr;
     char    len_buf[ORA_BUFF_SIZE];
     char    typ_buf[ORA_BUFF_SIZE];
+    char    prec_buf[ORA_BUFF_SIZE];
+    char    scale_buf[ORA_BUFF_SIZE];
     char    buf2[ORA_MSG_SIZE];
 
     if ((hand = ora_prologue(interp, argc, argv, 2, " cur_handle")) == -1) {
@@ -1424,9 +1683,11 @@ Tcl_OraCols (clientData, interp, argc, argv)
     }
 
 
-    len_buf[0] = '\0';   /* initial null buffers */
-    typ_buf[0] = '\0';   
-    buf2[0]    = '\0';   
+    len_buf[0]   = '\0';   /* initial null buffers */
+    typ_buf[0]   = '\0';   
+    prec_buf[0]  = '\0';
+    scale_buf[0] = '\0';
+    buf2[0]      = '\0';   
 
     for (col = OraCurs[hand].col_list; col != NULL; col = col->next_buf) {
 
@@ -1483,12 +1744,29 @@ Tcl_OraCols (clientData, interp, argc, argv)
 	sprintf(buf2, (i>1) ? " %s" : "%s",type_ptr);
 	strcat(typ_buf,buf2);
 
+        /* format precision and scale for numeric data */
+	if (col->dbtype == 2) {
+	  sprintf(buf2, (i>1) ? " %d" : "%d", col->prec);
+	} else {
+	  sprintf(buf2, (i>1) ? " %s" : "%s", "{}");
+	}
+	strcat(prec_buf,buf2);
+
+	if (col->dbtype == 2) {
+	  sprintf(buf2, (i>1) ? " %d" : "%d", col->scale);
+	} else {
+	  sprintf(buf2, (i>1) ? " %s" : "%s", "{}");
+	}
+	strcat(scale_buf,buf2);
+
 
     }
 
     if (i > 0) {
         Tcl_SetVar2(interp, OraMsgArray, "collengths", len_buf,TCL_GLOBAL_ONLY);
         Tcl_SetVar2(interp, OraMsgArray, "coltypes",   typ_buf,TCL_GLOBAL_ONLY);
+        Tcl_SetVar2(interp, OraMsgArray, "colprecs",  prec_buf,TCL_GLOBAL_ONLY);
+        Tcl_SetVar2(interp, OraMsgArray, "colscales",scale_buf,TCL_GLOBAL_ONLY);
     }
 
 
